from flex.datastructures import (
    ValidationDict,
)

from .multiple_of import (
    multiple_of_validator,
    validate_type_for_multiple_of,
)
from .maximum import (
    maximum_validator,
    exclusive_maximum_validator,
    validate_maximum_is_gte_minimum,
    validate_maximum_required_if_exclusive_maximum_set,
    validate_type_for_maximum,
)
from .minimum import (
    minimum_validator,
    exclusive_minimum_validator,
    validate_minimum_required_if_exclusive_minimum_set,
    validate_type_for_minimum,
)
from .min_length import (
    min_length_validator,
    validate_type_for_min_length,
)
from .max_length import (
    max_length_validator,
    validate_max_length_greater_than_or_equal_to_min_length,
    validate_type_for_max_length,
)
from .pattern import (
    pattern_validator,
)
from .min_items import (
    min_items_validator,
    validate_type_for_min_items,
)
from .max_items import (
    max_items_validator,
    validate_max_items_greater_than_or_equal_to_min_items,
    validate_type_for_max_items,
)
from .unique_items import (
    unique_items_validator,
    validate_type_for_unique_items,
)
from .enum import (
    enum_validator,
)


field_validators = ValidationDict()

field_validators.add_property_validator('multipleOf', multiple_of_validator)
field_validators.add_property_validator('minimum', minimum_validator)
field_validators.add_property_validator('maximum', maximum_validator)
field_validators.add_property_validator('exclusiveMinimum', exclusive_minimum_validator)
field_validators.add_property_validator('exclusiveMaximum', exclusive_maximum_validator)
field_validators.add_property_validator('minLength', min_length_validator)
field_validators.add_property_validator('maxLength', max_length_validator)
field_validators.add_property_validator('pattern', pattern_validator)
field_validators.add_property_validator('minItems', min_items_validator)
field_validators.add_property_validator('maxItems', max_items_validator)
field_validators.add_property_validator('uniqueItems', unique_items_validator)
field_validators.add_property_validator('enum', enum_validator)


non_field_validators = ValidationDict()

non_field_validators.add_validator(
    'maximum', validate_maximum_is_gte_minimum,
)
non_field_validators.add_validator(
    'maximum', validate_maximum_required_if_exclusive_maximum_set,
)
non_field_validators.add_validator(
    'minimum', validate_minimum_required_if_exclusive_minimum_set,
)
non_field_validators.add_validator(
    'maxLength', validate_max_length_greater_than_or_equal_to_min_length,
)
non_field_validators.add_validator(
    'maxItems', validate_max_items_greater_than_or_equal_to_min_items,
)


type_validators = ValidationDict()
type_validators.add_validator(
    'type', validate_type_for_multiple_of,
)
type_validators.add_validator(
    'type', validate_type_for_maximum,
)
type_validators.add_validator(
    'type', validate_type_for_minimum,
)
type_validators.add_validator(
    'type', validate_type_for_max_length,
)
type_validators.add_validator(
    'type', validate_type_for_min_length,
)
type_validators.add_validator(
    'type', validate_type_for_max_items,
)
type_validators.add_validator(
    'type', validate_type_for_min_items,
)
type_validators.add_validator(
    'type', validate_type_for_unique_items,
)



from flex.exceptions import ValidationError
from flex.constants import (
    INTEGER,
    NUMBER,
    EMPTY,
    BOOLEAN,
)
from flex.utils import (
    pluralize,
)
from flex.error_messages import MESSAGES
from flex.validation.common import (
    generate_object_validator,
)
from flex.decorators import (
    suffix_reserved_words,
    pull_keys_from_obj,
    skip_if_any_kwargs_empty,
)


@pull_keys_from_obj('minimum', 'exclusiveMinimum')
@skip_if_any_kwargs_empty('exclusiveMinimum')
def validate_minimum_required_if_exclusive_minimum_set(minimum, exclusiveMinimum, **kwargs):
    if exclusiveMinimum is True and minimum is EMPTY:
        raise ValidationError(
            MESSAGES['minimum']['exclusive_minimum_required_minimum'],
        )


@pull_keys_from_obj('type', 'minimum')
@suffix_reserved_words
@skip_if_any_kwargs_empty('type_', 'minimum')
def validate_type_for_minimum(type_, minimum, **kwargs):
    types = pluralize(type_)

    if not set(types).intersection((INTEGER, NUMBER)):
        raise ValidationError(
            MESSAGES['type']['invalid_type_for_minimum'],
        )


minimum_schema = {
    'type': NUMBER,
}
minimum_validator = generate_object_validator(
    schema=minimum_schema,
)

exclusive_minimum_schema = {
    'type': BOOLEAN,
}
exclusive_minimum_validator = generate_object_validator(
    schema=exclusive_minimum_schema,
)



from flex.exceptions import ValidationError
from flex.error_messages import MESSAGES
from flex.constants import (
    INTEGER,
    OBJECT,
)
from flex.utils import pluralize
from flex.validation.common import (
    generate_object_validator,
)
from flex.validation.schema import (
    construct_schema_validators,
)
from flex.decorators import (
    pull_keys_from_obj,
    suffix_reserved_words,
    skip_if_any_kwargs_empty,
)


@pull_keys_from_obj('type', 'minProperties')
@suffix_reserved_words
@skip_if_any_kwargs_empty('type_', 'minProperties')
def validate_type_for_min_properties(type_, minProperties, **kwargs):
    types = pluralize(type_)

    if not set(types).intersection((OBJECT,)):
        raise ValidationError(
            MESSAGES['type']['invalid_type_for_min_properties'],
        )


min_properties_schema = {
    'type': INTEGER,
    'minimum': 0,
}
min_properties_validators = construct_schema_validators(min_properties_schema, {})
min_properties_validator = generate_object_validator(
    field_validators=min_properties_validators,
)



from flex.datastructures import (
    ValidationDict,
)
from flex.error_messages import (
    MESSAGES,
)
from flex.constants import (
    OBJECT,
    ARRAY,
    EMPTY,
)
from flex.exceptions import ValidationError
from flex.utils import (
    pluralize,
)
from flex.decorators import (
    pull_keys_from_obj,
    suffix_reserved_words,
)
from flex.loading.common import (
    field_validators as common_field_validators,
    non_field_validators as common_non_field_validators,
    type_validators as common_type_validators,
)
from flex.loading.common.default import (
    validate_default_is_of_one_of_declared_types,
)
from flex.validation.common import (
    generate_object_validator,
)
from flex.loading.common.format import (
    format_validator,
)
from .description import (
    description_validator,
)
from .type import (
    type_validator,
)
from .collection_format import (
    collection_format_validator,
)


single_header_schema = {
    'type': OBJECT,
    'required': [
        'type',
    ]
}

single_header_field_validators = ValidationDict()
single_header_field_validators.update(common_field_validators)
single_header_field_validators.add_property_validator('description', description_validator)
single_header_field_validators.add_property_validator('type', type_validator)
single_header_field_validators.add_property_validator('format', format_validator)
single_header_field_validators.add_property_validator(
    'collectionFormat', collection_format_validator,
)


@pull_keys_from_obj('type', 'items')
@suffix_reserved_words
def validate_items_required_if_type_arraw(type_, items, **kwargs):
    types = pluralize(type_)
    if ARRAY in types and items is EMPTY:
        raise ValidationError(MESSAGES['required']['required'])


single_header_non_field_validators = ValidationDict()
single_header_non_field_validators.update(common_non_field_validators)
single_header_non_field_validators.update(common_type_validators)
single_header_non_field_validators.add_validator(
    'default', validate_default_is_of_one_of_declared_types,
)
single_header_non_field_validators.add_validator(
    'items', validate_items_required_if_type_arraw,
)

single_header_validator = generate_object_validator(
    schema=single_header_schema,
    field_validators=single_header_field_validators,
    non_field_validators=single_header_non_field_validators,
)



from six.moves import urllib_parse as urlparse

from flex.datastructures import (
    ValidationList,
)
from flex.constants import (
    STRING,
)
from flex.exceptions import ValidationError
from flex.error_messages import MESSAGES
from flex.validation.common import (
    generate_object_validator,
    generate_type_validator,
)
from flex.decorators import (
    skip_if_empty,
    skip_if_not_of_type,
)


string_type_validator = generate_type_validator(STRING)


@skip_if_empty
@skip_if_not_of_type(STRING)
def base_path_validator(value, **kwargs):
    if not value.startswith('/'):
        raise ValidationError(MESSAGES['path']['must_start_with_slash'])
    parts = urlparse.urlparse(value)
    if value != parts.path:
        raise ValidationError(MESSAGES['path']['invalid'])


base_path_schema = {
    'type': STRING,
}

non_field_validators = ValidationList()
non_field_validators.add_validator(base_path_validator)

base_path_validator = generate_object_validator(
    schema=base_path_schema,
    non_field_validators=non_field_validators,
)



from flex.constants import (
    BOOLEAN,
)
from flex.validation.common import (
    generate_object_validator,
)
from flex.validation.schema import (
    construct_schema_validators,
)


required_schema = {
    'type': BOOLEAN,
}
required_validators = construct_schema_validators(required_schema, {})

required_validator = generate_object_validator(
    field_validators=required_validators,
)



from flex.exceptions import ValidationError
from flex.error_messages import MESSAGES
from flex.constants import (
    INTEGER,
    EMPTY,
    ARRAY,
)
from flex.utils import pluralize
from flex.validation.common import (
    generate_object_validator,
)
from flex.validation.schema import (
    construct_schema_validators,
)
from flex.decorators import (
    pull_keys_from_obj,
    suffix_reserved_words,
    skip_if_any_kwargs_empty,
)


@pull_keys_from_obj('minItems', 'maxItems')
def validate_max_items_greater_than_or_equal_to_min_items(minItems, maxItems, **kwargs):
    if minItems is EMPTY or maxItems is EMPTY:
        return
    if not maxItems >= minItems:
        raise ValidationError(
            MESSAGES['max_items']['must_be_greater_than_min_items']
        )


@pull_keys_from_obj('type', 'maxItems')
@suffix_reserved_words
@skip_if_any_kwargs_empty('type_', 'maxItems')
def validate_type_for_max_items(type_, maxItems, **kwargs):
    types = pluralize(type_)

    if not set(types).intersection((ARRAY,)):
        raise ValidationError(
            MESSAGES['type']['invalid_type_for_max_items'],
        )


max_items_schema = {
    'type': INTEGER,
}
max_items_validators = construct_schema_validators(max_items_schema, {})

max_items_validator = generate_object_validator(
    field_validators=max_items_validators,
)



import re

from flex.datastructures import (
    ValidationList,
)
from flex.constants import (
    STRING,
)
from flex.error_messages import MESSAGES
from flex.validation.common import (
    generate_object_validator,
    generate_type_validator,
)
from flex.decorators import (
    skip_if_empty,
    skip_if_not_of_type,
)
from flex.context_managers import ErrorCollection


string_type_validator = generate_type_validator(STRING)


def decompose_hostname(value):
    scheme, _, right = value.rpartition('://')
    left, _, right = right.partition(':')
    if right:
        hostname = left
        port, _, path = right.partition('/')
    else:
        port = ''
        hostname, _, path = left.partition('/')

    return scheme, hostname, port, path


@skip_if_empty
@skip_if_not_of_type(STRING)
def host_validator(value, **kwargs):
    """
    From: http://stackoverflow.com/questions/2532053/validate-a-hostname-string
    According to: http://en.wikipedia.org/wiki/Hostname#Restrictions_on_valid_host_names
    """
    scheme, hostname, port, path = decompose_hostname(value)

    if len(hostname) > 255:
        return False
    if hostname[-1] == ".":
        hostname = hostname[:-1]  # strip exactly one dot from the right, if present
    allowed = re.compile("(?!-)[A-Z\d-]{1,63}(?<!-)$", re.IGNORECASE)

    with ErrorCollection() as errors:
        if not all(allowed.match(x) for x in hostname.split(".")):
            errors.add_error(
                'invalid',
                MESSAGES['host']['invalid'].format(value),
            )

        if path:
            errors.add_error(
                'path',
                MESSAGES['host']['may_not_include_path'].format(value),
            )

        if scheme:
            errors.add_error(
                'scheme',
                MESSAGES['host']['may_not_include_scheme'].format(value),
            )


host_schema = {
    'type': STRING,
}

non_field_validators = ValidationList()
non_field_validators.add_validator(host_validator)

host_validator = generate_object_validator(
    schema=host_schema,
    non_field_validators=non_field_validators,
)



from flex.exceptions import ValidationError
from flex.error_messages import MESSAGES
from flex.constants import (
    STRING,
    MULTI,
    QUERY,
    FORM_DATA,
    COLLECTION_FORMATS,
)
from flex.validation.common import (
    generate_object_validator,
)
from flex.decorators import (
    pull_keys_from_obj,
    suffix_reserved_words,
    skip_if_any_kwargs_empty,
)


@pull_keys_from_obj('in', 'collectionFormat')
@skip_if_any_kwargs_empty('in', 'collectionFormat')
@suffix_reserved_words
def validate_collection_format_based_on_in_value(in_, collectionFormat, **kwargs):
    if collectionFormat == MULTI:
        if in_ not in (QUERY, FORM_DATA):
            raise ValidationError(MESSAGES['collection_format']['invalid_based_on_in_value'])


collection_format_schema = {
    'type': STRING,
    'enum': COLLECTION_FORMATS,
}


collection_format_validator = generate_object_validator(
    schema=collection_format_schema,
)



from flex.datastructures import (
    ValidationDict,
)
from flex.constants import (
    OBJECT,
    STRING,
)
from flex.validation.common import (
    generate_object_validator,
)
from .headers import (
    headers_validator,
)
from .schema import (
    schema_validator,
)


single_response_schema = {
    'type': OBJECT,
    'required': [
        'description',
    ],
    'properties': {
        'description': {
            'type': STRING,
        },
    },
}


field_validators = ValidationDict()

field_validators.add_property_validator('schema', schema_validator)
field_validators.add_property_validator('headers', headers_validator)


single_response_validator = generate_object_validator(
    schema=single_response_schema,
    field_validators=field_validators,
)



from flex.error_messages import MESSAGES
from flex.exceptions import ValidationError
from flex.constants import (
    INTEGER,
    STRING,
)
from flex.utils import pluralize
from flex.validation.common import (
    generate_object_validator,
)
from flex.decorators import (
    pull_keys_from_obj,
    suffix_reserved_words,
    skip_if_any_kwargs_empty,
)


@pull_keys_from_obj('type', 'minLength')
@suffix_reserved_words
@skip_if_any_kwargs_empty('type_', 'minLength')
def validate_type_for_min_length(type_, minLength, **kwargs):
    types = pluralize(type_)

    if not set(types).intersection((STRING,)):
        raise ValidationError(
            MESSAGES['type']['invalid_type_for_min_length'],
        )


min_length_schema = {
    'type': INTEGER,
    'minimum': 0,
}
min_length_validator = generate_object_validator(
    schema=min_length_schema,
)



from flex.exceptions import ValidationError
from flex.error_messages import MESSAGES
from flex.constants import (
    STRING,
    PARAMETER_IN_VALUES,
    PATH,
    BODY,
    EMPTY,
)
from flex.validation.common import (
    generate_object_validator,
)
from flex.decorators import (
    pull_keys_from_obj,
    suffix_reserved_words,
    skip_if_any_kwargs_empty,
)


in_schema = {
    'type': STRING,
    'enum': PARAMETER_IN_VALUES,
}


@pull_keys_from_obj('in', 'required')
@skip_if_any_kwargs_empty('in')
@suffix_reserved_words
def validate_path_parameters_must_be_required(in_, required, **kwargs):
    if in_ == PATH:
        if required is not True:
            raise ValidationError(MESSAGES['required']['path_parameters_must_be_required'])


@pull_keys_from_obj('in', 'schema')
@skip_if_any_kwargs_empty('in')
@suffix_reserved_words
def validate_body_parameters_must_include_a_schema(in_, schema, **kwargs):
    if in_ == BODY:
        if schema is EMPTY:
            raise ValidationError(MESSAGES['schema']['body_parameters_must_include_a_schema'])


@pull_keys_from_obj('in', 'type')
@skip_if_any_kwargs_empty('in')
@suffix_reserved_words
def validate_type_declared_for_non_body_parameters(in_, type_, **kwargs):
    if in_ != BODY:
        if type_ is EMPTY:
            raise ValidationError(MESSAGES['type']['non_body_parameters_must_declare_a_type'])


in_validator = generate_object_validator(
    schema=in_schema,
)



from flex.exceptions import (
    ValidationError,
)
from flex.error_messages import (
    MESSAGES,
)
from flex.constants import (
    NUMBER,
    INTEGER,
)
from flex.validation.common import (
    generate_object_validator,
)
from flex.utils import (
    pluralize,
)
from flex.decorators import (
    pull_keys_from_obj,
    suffix_reserved_words,
    skip_if_any_kwargs_empty,
)


multiple_of_schema = {
    'type': NUMBER,
    'minimum': 0,
}

multiple_of_validator = generate_object_validator(
    schema=multiple_of_schema,
)


@pull_keys_from_obj('type', 'multipleOf')
@suffix_reserved_words
@skip_if_any_kwargs_empty('type_', 'multipleOf')
def validate_type_for_multiple_of(type_, multipleOf, **kwargs):
    types = pluralize(type_)

    if not set(types).intersection((INTEGER, NUMBER)):
        raise ValidationError(
            MESSAGES['type']['invalid_type_for_multiple_of'],
        )



import functools

from flex.constants import (
    ARRAY,
    OBJECT,
)
from flex.decorators import (
    skip_if_not_of_type,
    skip_if_empty,
)
from flex.validation.common import (
    generate_object_validator,
    apply_validator_to_object,
    apply_validator_to_array,
)
from flex.datastructures import (
    ValidationDict,
    ValidationList,
)
from .ref import (
    ref_validator,
)
from flex.loading.common.schema import (
    schema_schema as common_schema_schema,
    schema_field_validators as common_schema_field_validators,
    schema_non_field_validators as common_schema_non_field_validators,
    properties_schema as common_properties_schema,
    items_schema as common_items_schema,
)

schema_field_validators = ValidationDict()
schema_field_validators.update(common_schema_field_validators)

schema_field_validators.add_property_validator('$ref', ref_validator)

schema_non_field_validators = ValidationDict()
schema_non_field_validators.update(common_schema_non_field_validators)


schema_validator = generate_object_validator(
    schema=common_schema_schema,
    field_validators=schema_field_validators,
    non_field_validators=schema_non_field_validators,
)


#
# Properties.
#
properties_non_field_validators = ValidationDict()
properties_non_field_validators.add_validator(
    'properties', functools.partial(apply_validator_to_object, validator=schema_validator),
)

properties_validator = generate_object_validator(
    schema=common_properties_schema,
    non_field_validators=properties_non_field_validators,
)

# Now put the properties validator onto the schema validator.
schema_field_validators.add_property_validator('properties', properties_validator)


#
# Items
#
items_non_field_validators = ValidationList()
items_non_field_validators.add_validator(
    skip_if_empty(skip_if_not_of_type(OBJECT)(schema_validator))
)
items_non_field_validators.add_validator(
    skip_if_empty(skip_if_not_of_type(ARRAY)(
        functools.partial(apply_validator_to_array, validator=schema_validator),
    )),
)

items_validator = generate_object_validator(
    schema=common_items_schema,
    non_field_validators=items_non_field_validators,
)

# Now put the items validator onto the schema validator
schema_field_validators.add_property_validator('items', items_validator)



from __future__ import unicode_literals

from six.moves import urllib_parse as urlparse
import os
import collections
import requests

import six
import json
import yaml

from flex.context_managers import ErrorCollection
from flex.exceptions import ValidationError
from flex.loading.definitions import (
    definitions_validator,
)
from flex.loading.schema import (
    swagger_schema_validator,
)
from flex.loading.schema.paths.path_item.operation.responses.single.schema import (
    schema_validator,
)
from flex.http import (
    normalize_request,
    normalize_response,
)
from flex.validation.common import validate_object
from flex.validation.request import validate_request
from flex.validation.response import validate_response


def load_source(source):
    """
    Common entry point for loading some form of raw swagger schema.

    Supports:
        - python object (dictionary-like)
        - path to yaml file
        - path to json file
        - file object (json or yaml).
        - json string.
        - yaml string.
    """
    if isinstance(source, collections.Mapping):
        return source
    elif hasattr(source, 'read') and callable(source.read):
        raw_source = source.read()
    elif os.path.exists(os.path.expanduser(str(source))):
        with open(os.path.expanduser(str(source)), 'r') as source_file:
            raw_source = source_file.read()
    elif isinstance(source, six.string_types):
        parts = urlparse.urlparse(source)
        if parts.scheme and parts.netloc:
            response = requests.get(source)
            if isinstance(response.content, six.binary_type):
                raw_source = six.text_type(response.content, encoding='utf-8')
            else:
                raw_source = response.content
        else:
            raw_source = source

    try:
        try:
            return json.loads(raw_source)
        except ValueError:
            pass

        try:
            return yaml.load(raw_source)
        except (yaml.scanner.ScannerError, yaml.parser.ParserError):
            pass
    except NameError:
        pass

    raise ValueError(
        "Unable to parse `{0}`.  Tried yaml and json.".format(source),
    )


def parse(raw_schema):
    context = {
        'deferred_references': set(),
    }
    swagger_definitions = definitions_validator(raw_schema, context=context)

    swagger_schema = swagger_schema_validator(
        raw_schema,
        context=swagger_definitions,
    )
    return swagger_schema


def load(target):
    """
    Given one of the supported target formats, load a swagger schema into it's
    python representation.
    """
    raw_schema = load_source(target)
    return parse(raw_schema)


def validate(raw_schema, target=None, **kwargs):
    """
    Given the python representation of a JSONschema as defined in the swagger
    spec, validate that the schema complies to spec.  If `target` is provided,
    that target will be validated against the provided schema.
    """
    schema = schema_validator(raw_schema)

    if target is not None:
        validate_object(target, schema=schema)


def validate_api_call(schema, raw_request, raw_response):
    """
    Validate the request/response cycle of an api call against a swagger
    schema.  Request/Response objects from the `requests` and `urllib` library
    are supported.
    """
    request = normalize_request(raw_request)
    response = normalize_response(raw_response)

    with ErrorCollection() as errors:
        try:
            validate_request(
                request=request,
                schema=schema,
            )
        except ValidationError as err:
            errors['request'].add_error(err.messages or getattr(err, 'detail'))
            return

        try:
            validate_response(
                response=response,
                request_method=request.method,
                schema=schema,
            )
        except ValidationError as err:
            errors['response'].add_error(err.messages or getattr(err, 'detail'))



from flex.constants import (
    ARRAY,
)
from flex.validation.common import (
    generate_object_validator,
)
from flex.validation.schema import (
    construct_schema_validators,
)


enum_schema = {
    'type': ARRAY,
    'minItems': 1,
}
enum_validators = construct_schema_validators(enum_schema, {})

enum_validator = generate_object_validator(
    field_validators=enum_validators,
)



from flex.constants import (
    STRING,
)
from flex.datastructures import (
    ValidationList,
)
from flex.validation.common import (
    generate_object_validator,
)
from flex.decorators import (
    skip_if_not_of_type,
    skip_if_empty,
)


@skip_if_empty
@skip_if_not_of_type(STRING)
def defer_reference(reference, context, **kwargs):
    try:
        deferred_references = context['deferred_references']
    except KeyError:
        raise KeyError("`deferred_references` not found in validation context")
    deferred_references.add(reference)


ref_schema = {
    'type': STRING,
}

non_field_validators = ValidationList()
non_field_validators.add_validator(defer_reference)

ref_validator = generate_object_validator(
    schema=ref_schema,
    non_field_validators=non_field_validators,
)



from flex.constants import (
    STRING,
    CSV,
    COLLECTION_FORMATS,
)
from flex.validation.common import (
    generate_object_validator,
)


collection_format_schema = {
    'type': STRING,
    'default': CSV,
    'enum': COLLECTION_FORMATS,
}


collection_format_validator = generate_object_validator(
    schema=collection_format_schema,
)



from flex.exceptions import (
    ValidationError,
    ErrorList,
)
from flex.constants import (
    NULL,
    BOOLEAN,
    INTEGER,
    NUMBER,
    STRING,
    ARRAY,
    OBJECT,
)
from flex.utils import (
    pluralize,
)
from flex.validation.common import (
    generate_object_validator,
)
from flex.decorators import (
    suffix_reserved_words,
    skip_if_not_of_type,
)
from flex.validation.schema import (
    construct_schema_validators,
)


single_type_schema = {
    'type': STRING,
    'enum': [
        NULL,
        BOOLEAN,
        INTEGER,
        NUMBER,
        STRING,
        ARRAY,
        OBJECT,
    ],
}
single_type_validators = construct_schema_validators(single_type_schema, {})

single_type_validator = generate_object_validator(
    field_validators=single_type_validators,
)


@suffix_reserved_words
@skip_if_not_of_type(ARRAY, STRING)
def validate_types(type_, **kwargs):
    types = pluralize(type_)

    with ErrorList() as errors:
        for value in types:
            try:
                single_type_validator(value)
            except ValidationError as err:
                errors.add_error(err.detail)


type_schema = {
    'type': [
        STRING,
        ARRAY,
    ],
}
type_validators = construct_schema_validators(type_schema, {})
type_validators.add_validator('type', validate_types)
type_validator = generate_object_validator(
    field_validators=type_validators,
)



from flex.datastructures import (
    ValidationList,
)
from flex.exceptions import (
    ValidationError,
    ErrorDict,
)
from flex.constants import OBJECT
from flex.decorators import (
    skip_if_empty,
    skip_if_not_of_type,
)
from flex.validation.common import (
    generate_object_validator,
)
from .schema import (
    schema_validator,
)


@skip_if_empty
@skip_if_not_of_type(OBJECT)
def validate_schema_definitions(definitions, **kwargs):
    with ErrorDict() as errors:
        for name, schema in definitions.items():
            try:
                schema_validator(schema, **kwargs)
            except ValidationError as err:
                errors.add_error(name, err.detail)


schema_definitions_schema = {
    'type': OBJECT,
}

non_field_validators = ValidationList()
non_field_validators.add_validator(validate_schema_definitions)

schema_definitions_validator = generate_object_validator(
    schema=schema_definitions_schema,
    non_field_validators=non_field_validators,
)



from flex.datastructures import (
    ValidationDict,
)
from flex.constants import (
    BOOLEAN,
    ARRAY,
    STRING,
    OBJECT,
)
from flex.validation.common import (
    generate_object_validator,
)
from flex.loading.common.mimetypes import (
    mimetype_validator,
)
from flex.loading.common.external_docs import (
    external_docs_validator,
)
from .parameters import (
    parameters_validator,
)
from .responses import (
    responses_validator,
)


operation_schema = {
    'type': OBJECT,
    'properties': {
        'tags': {
            'type': ARRAY,
            'items': {
                'type': STRING,
            }
        },
        'summary': {
            'type': STRING,
        },
        'description': {
            'type': STRING,
        },
        'operationId': {
            'type': STRING,
        },
        'consumes': {
            'type': ARRAY,
            'items': {
                'type': STRING,
            },
        },
        'produces': {
            'type': ARRAY,
            'items': {
                'type': STRING,
            },
        },
        # 'schemes':  # TODO:  Implement this.
        'deprecated': {
            'type': BOOLEAN,
        },
        # 'security':  # TODO:  Implement this.
    },
}


field_validators = ValidationDict()
field_validators.add_property_validator('parameters', parameters_validator)
field_validators.add_property_validator('responses', responses_validator)
field_validators.add_property_validator('externalDocs', external_docs_validator)

non_field_validators = ValidationDict()
non_field_validators.add_property_validator('consumes', mimetype_validator)
non_field_validators.add_property_validator('produces', mimetype_validator)

operation_validator = generate_object_validator(
    schema=operation_schema,
    field_validators=field_validators,
)



from flex.exceptions import (
    ErrorDict,
)


class ErrorCollection(ErrorDict):
    def __init__(self, *args, **kwargs):
        import warnings
        warnings.warn("user `flex.exceptions.ErrorDict` instead", DeprecationWarning)
        super(ErrorCollection, self).__init__(*args, **kwargs)



from flex.constants import (
    ARRAY,
    SCHEMES,
)
from flex.exceptions import ValidationError
from flex.datastructures import (
    ValidationList,
)
from flex.error_messages import MESSAGES
from flex.validation.common import (
    generate_object_validator,
)
from flex.decorators import (
    skip_if_empty,
    skip_if_not_of_type,
)


@skip_if_empty
@skip_if_not_of_type(ARRAY)
def validate_schemes(schemes, **kwargs):
    for value in schemes:
        if value not in SCHEMES:
            raise ValidationError(
                MESSAGES['schemes']['invalid'].format(value),
            )


scheme_schema = {
    'type': ARRAY,
}

non_field_validators = ValidationList()
non_field_validators.add_validator(validate_schemes)

schemes_validator = generate_object_validator(
    schema=scheme_schema,
    non_field_validators=non_field_validators,
)



from flex.datastructures import (
    ValidationList,
)
from flex.exceptions import ValidationError
from flex.constants import (
    OBJECT,
)
from flex.error_messages import MESSAGES
from flex.decorators import (
    skip_if_not_of_type,
    skip_if_empty,
)
from flex.validation.common import (
    generate_object_validator,
)
from flex.context_managers import (
    ErrorCollection,
)
from .path_item import (
    path_item_validator,
)


@skip_if_empty
@skip_if_not_of_type(OBJECT)
def validate_path_items(paths, **kwargs):
    with ErrorCollection() as errors:
        for path, path_definition in paths.items():
            # TODO: move this to its own validation function that validates the keys.
            if not path.startswith('/'):
                errors.add_error(path, MESSAGES['path']['must_start_with_slash'])

            try:
                path_item_validator(path_definition, **kwargs)
            except ValidationError as err:
                errors.add_error(path, err.detail)


paths_schema = {
    'type': OBJECT,
}
non_field_validators = ValidationList()
non_field_validators.add_validator(validate_path_items)

paths_validator = generate_object_validator(
    schema=paths_schema,
    non_field_validators=non_field_validators,
)



from flex.constants import (
    OBJECT,
    URI,
    STRING,
)
from flex.validation.common import (
    generate_object_validator,
)


external_docs_schema = {
    'type': OBJECT,
    'required': [
        'url',
    ],
    'properties': {
        'description': {
            'type': STRING,
        },
        'url': {
            'type': STRING,
            'format': URI,
        },
    },
}
external_docs_validator = generate_object_validator(
    schema=external_docs_schema,
)



import six

from flex.exceptions import ValidationError
from flex.error_messages import MESSAGES
from flex.decorators import (
    pull_keys_from_obj,
    suffix_reserved_words,
    skip_if_any_kwargs_empty,
)
from flex.utils import is_value_of_any_type


@pull_keys_from_obj('type', 'default')
@suffix_reserved_words
@skip_if_any_kwargs_empty('default', 'type_')
def validate_default_is_of_one_of_declared_types(default, type_, **kwargs):
    if isinstance(type_, six.string_types):
        type_ = [type_]

    if not is_value_of_any_type(default, type_):
        raise ValidationError(
            MESSAGES['default']['invalid_type'].format(default, type_),
        )



from flex.datastructures import (
    ValidationDict,
)
from flex.constants import (
    OBJECT,
)
from flex.validation.common import (
    generate_object_validator,
)
from .operation import (
    operation_validator,
)


def parameters_validator(*args, **kwargs):
    pass


path_item_schema = {
    'type': OBJECT,
}

non_field_validators = ValidationDict()
non_field_validators.add_property_validator('get', operation_validator)
non_field_validators.add_property_validator('put', operation_validator)
non_field_validators.add_property_validator('post', operation_validator)
non_field_validators.add_property_validator('delete', operation_validator)
non_field_validators.add_property_validator('options', operation_validator)
non_field_validators.add_property_validator('head', operation_validator)
non_field_validators.add_property_validator('patch', operation_validator)
non_field_validators.add_property_validator('parameters', parameters_validator)

path_item_validator = generate_object_validator(
    schema=path_item_schema,
    non_field_validators=non_field_validators,
)



import operator
import functools

from flex.constants import EMPTY


def chain_reduce(value, functions, **kwargs):
    for function in functions:
        value = function(value, **kwargs)
    return value


def chain_reduce_partial(*functions):
    """
    Given an iterable of functions, returns a callable that takes a value and
    passes it through all of the given functions in order.

    def a(x):
        ...

    def b(x):
        ...

    c = chain_reduce_partial(a, b)

    This is equivilent to

    def c(x):
        return b(a(x))
    """
    return functools.partial(
        chain_reduce,
        functions=functions,
    )


def apply_functions_to_key(key, *funcs):
    """
    Shortcut for the common pattern with `chain_reduce_partial` for applying a
    validator to some specified key in a mapping.
    """
    return chain_reduce_partial(
        operator.methodcaller('get', key, EMPTY), *funcs
    )



from flex.exceptions import ValidationError
from flex.error_messages import MESSAGES
from flex.constants import (
    INTEGER,
    NUMBER,
    EMPTY,
    BOOLEAN
)
from flex.utils import (
    pluralize,
)
from flex.validation.common import (
    generate_object_validator,
)
from flex.decorators import (
    pull_keys_from_obj,
    suffix_reserved_words,
    skip_if_any_kwargs_empty,
)


@pull_keys_from_obj('minimum', 'maximum')
def validate_maximum_is_gte_minimum(minimum, maximum, **kwargs):
    if minimum is EMPTY or maximum is EMPTY:
        return
    if not maximum >= minimum:
        raise ValidationError(MESSAGES['maximum']['must_be_greater_than_minimum'])


@pull_keys_from_obj('maximum', 'exclusiveMaximum')
def validate_maximum_required_if_exclusive_maximum_set(maximum, exclusiveMaximum, **kwargs):
    if exclusiveMaximum is EMPTY:
        return
    if exclusiveMaximum is True and maximum is EMPTY:
        raise ValidationError(
            MESSAGES['maximum']['exclusive_maximum_required_maximum'],
        )


@pull_keys_from_obj('type', 'maximum')
@suffix_reserved_words
@skip_if_any_kwargs_empty('type_', 'maximum')
def validate_type_for_maximum(type_, maximum, **kwargs):
    types = pluralize(type_)

    if not set(types).intersection((INTEGER, NUMBER)):
        raise ValidationError(
            MESSAGES['type']['invalid_type_for_maximum'],
        )


maximum_schema = {
    'type': NUMBER,
}
maximum_validator = generate_object_validator(
    schema=maximum_schema,
)

exclusive_maximum_schema = {
    'type': BOOLEAN,
}
exclusive_maximum_validator = generate_object_validator(
    schema=exclusive_maximum_schema,
)



from flex.exceptions import ValidationError
from flex.error_messages import MESSAGES
from flex.constants import (
    ARRAY,
    EMPTY,
)
from flex.utils import (
    pluralize,
)
from flex.decorators import (
    pull_keys_from_obj,
    suffix_reserved_words,
    skip_if_any_kwargs_empty,
)


@pull_keys_from_obj('type', 'items')
@skip_if_any_kwargs_empty('type')
@suffix_reserved_words
def validate_items_required_if_array_type(type_, items, **kwargs):
    types = pluralize(type_)

    if ARRAY in types and items is EMPTY:
        raise ValidationError(MESSAGES['items']['items_required_for_type_array'])



from flex.exceptions import ValidationError
from flex.error_messages import MESSAGES
from flex.datastructures import (
    ValidationList,
)
from flex.constants import (
    STRING,
    ARRAY,
    NUMBER,
    INTEGER,
    BOOLEAN,
)
from flex.decorators import (
    skip_if_empty,
    skip_if_not_of_type,
)
from flex.validation.common import (
    generate_object_validator,
)

ALLOWED_TYPES = [STRING, NUMBER, INTEGER, BOOLEAN, ARRAY]

type_schema = {
    'type': [STRING, ARRAY],
    'items': {
        'type': STRING,
        'enum': ALLOWED_TYPES,
    },
}


@skip_if_not_of_type(STRING)
@skip_if_empty
def validate_type(_type, **kwargs):
    if _type not in ALLOWED_TYPES:
        raise ValidationError(MESSAGES['enum']['invalid'].format(_type, ALLOWED_TYPES))


non_field_validators = ValidationList()
non_field_validators.add_validator(validate_type)


type_validator = generate_object_validator(
    schema=type_schema,
    non_field_validators=non_field_validators,
)



from flex.datastructures import (
    ValidationDict,
)
from flex.constants import (
    OBJECT,
)
from flex.loading.common import (
    field_validators as common_field_validators,
    non_field_validators as common_non_field_validators,
    type_validators as common_type_validators,
)

from .in_ import (
    in_validator,
    validate_path_parameters_must_be_required,
    validate_body_parameters_must_include_a_schema,
    validate_type_declared_for_non_body_parameters,
)
from .name import (
    name_validator,
)
from .description import (
    description_validator,
)
from flex.loading.common.required import (
    required_validator,
)
from .type import (
    type_validator,
)
from flex.loading.common.format import (
    format_validator,
)
from .collection_format import (
    collection_format_validator,
    validate_collection_format_based_on_in_value,
)
from flex.loading.common.default import (
    validate_default_is_of_one_of_declared_types,
)
from .items import (
    validate_items_required_if_array_type,
)


single_parameter_schema = {
    'type': OBJECT,
    'required': [
        'name',
        'in',
    ],
}

single_parameter_field_validators = ValidationDict()
single_parameter_field_validators.update(common_field_validators)

single_parameter_field_validators.add_property_validator('in', in_validator)
single_parameter_field_validators.add_property_validator('name', name_validator)
single_parameter_field_validators.add_property_validator('description', description_validator)
single_parameter_field_validators.add_property_validator('required', required_validator)
single_parameter_field_validators.add_property_validator('type', type_validator)
single_parameter_field_validators.add_property_validator('format', format_validator)
single_parameter_field_validators.add_property_validator(
    'collectionFormat', collection_format_validator,
)

single_parameter_non_field_validators = ValidationDict()
single_parameter_non_field_validators.update(common_non_field_validators)
single_parameter_non_field_validators.update(common_type_validators)
single_parameter_non_field_validators.add_validator(
    'default', validate_default_is_of_one_of_declared_types,
)
single_parameter_non_field_validators.add_validator(
    'required', validate_path_parameters_must_be_required,
)
single_parameter_non_field_validators.add_validator(
    'schema', validate_body_parameters_must_include_a_schema,
)
single_parameter_non_field_validators.add_validator(
    'type', validate_type_declared_for_non_body_parameters,
)
single_parameter_non_field_validators.add_validator(
    'collectionFormat', validate_collection_format_based_on_in_value,
)
single_parameter_non_field_validators.add_validator(
    'items', validate_items_required_if_array_type,
)



from flex.exceptions import ErrorDict
from flex.error_messages import MESSAGES
from flex.datastructures import (
    ValidationDict,
)
from flex.constants import (
    OBJECT,
)
from flex.validation.common import (
    generate_object_validator,
)
from flex.decorators import (
    skip_if_not_of_type,
    pull_keys_from_obj,
)

from .schema_definitions import schema_definitions_validator
from .parameters import (
    parameters_validator,
)
from .responses import (
    responses_validator,
)


__ALL__ = [
    'schema_definitions_validator',
    'parameters_validator',
    'responses_validator',
]

definitions_schema = {
    'type': OBJECT,
}


@skip_if_not_of_type(OBJECT)
@pull_keys_from_obj('definitions')
def validate_references(definitions, context, **kwargs):
    try:
        deferred_references = context['deferred_references']
    except:
        raise KeyError("`deferred_references` not found in context")

    with ErrorDict() as errors:
        for reference in deferred_references:
            if reference not in definitions:
                errors.add_error(
                    reference,
                    MESSAGES['reference']['undefined'].format(reference),
                )


field_validators = ValidationDict()
field_validators.add_property_validator('definitions', schema_definitions_validator)
field_validators.add_property_validator('parameters', parameters_validator)
field_validators.add_property_validator('responses', responses_validator)

non_field_validators = ValidationDict()
non_field_validators.add_validator('definitions', validate_references)


definitions_validator = generate_object_validator(
    schema=definitions_schema,
    field_validators=field_validators,
    non_field_validators=non_field_validators,
)



import functools

from flex.datastructures import ValidationList
from flex.constants import (
    OBJECT,
)
from flex.validation.common import (
    generate_object_validator,
    apply_validator_to_object,
)

from .single import (
    single_header_validator,
)


headers_schema = {
    'type': OBJECT,
}


non_field_validators = ValidationList()
non_field_validators.add_validator(
    functools.partial(apply_validator_to_object, validator=single_header_validator),
)


headers_validator = generate_object_validator(
    schema=headers_schema,
    non_field_validators=non_field_validators,
)



from flex.datastructures import (
    ValidationDict,
)
from flex.loading.common.single_parameter import (
    single_parameter_schema,
    single_parameter_field_validators as common_single_parameter_field_validators,
    single_parameter_non_field_validators as common_single_parameter_non_field_validators,
)
from flex.loading.definitions.schema import (
    schema_validator,
    items_validator,
)
from flex.validation.common import (
    generate_object_validator,
)


single_parameter_field_validators = ValidationDict()
single_parameter_field_validators.update(
    common_single_parameter_field_validators
)

# schema fields
single_parameter_field_validators.add_property_validator('schema', schema_validator)
single_parameter_field_validators.add_property_validator('items', items_validator)

single_parameter_non_field_validators = ValidationDict()
single_parameter_non_field_validators.update(
    common_single_parameter_non_field_validators
)

single_parameter_validator = generate_object_validator(
    schema=single_parameter_schema,
    field_validators=single_parameter_field_validators,
    non_field_validators=single_parameter_non_field_validators,
)



from flex.exceptions import ValidationError
from flex.error_messages import MESSAGES
from flex.constants import (
    INTEGER,
    EMPTY,
    OBJECT,
)
from flex.utils import pluralize
from flex.validation.common import (
    generate_object_validator,
)
from flex.validation.schema import (
    construct_schema_validators,
)
from flex.decorators import (
    pull_keys_from_obj,
    suffix_reserved_words,
    skip_if_any_kwargs_empty,
)


@pull_keys_from_obj('minProperties', 'maxProperties')
def validate_max_properties_is_greater_than_or_equal_to_min_properties(minProperties,
                                                                       maxProperties,
                                                                       **kwargs):
    if maxProperties is EMPTY or minProperties is EMPTY:
        return
    if not maxProperties >= minProperties:
        raise ValidationError(
            MESSAGES['max_properties']['must_be_greater_than_min_properties'],
        )


@pull_keys_from_obj('type', 'maxProperties')
@suffix_reserved_words
@skip_if_any_kwargs_empty('type_', 'maxProperties')
def validate_type_for_max_properties(type_, maxProperties, **kwargs):
    types = pluralize(type_)

    if not set(types).intersection((OBJECT,)):
        raise ValidationError(
            MESSAGES['type']['invalid_type_for_max_properties'],
        )


max_properties_schema = {
    'type': INTEGER,
    'minimum': 0,
}
max_properties_validators = construct_schema_validators(max_properties_schema, {})
max_properties_validator = generate_object_validator(
    field_validators=max_properties_validators,
)



from flex.constants import (
    OBJECT,
)
from flex.datastructures import (
    ValidationDict,
)
from flex.validation.common import (
    generate_object_validator,
)
from flex.loading.common.mimetypes import (
    mimetype_validator,
)
from .info import info_validator
from .swagger import swagger_version_validator
from .host import host_validator
from .base_path import base_path_validator
from .schemes import schemes_validator
from .paths import paths_validator


__ALL__ = [
    'info_validator',
    'swagger_schema_validators',
    'host_validator',
    'base_path_validator',
    'schemes_validator',
    'mimetype_validator',
    'paths_validator',
]

swagger_schema = {
    'type': OBJECT,
    'required': [
        'info',
        'paths',
        'swagger',
    ],
}

non_field_validators = ValidationDict()
non_field_validators.add_property_validator('info', info_validator)
non_field_validators.add_property_validator('swagger', swagger_version_validator)
non_field_validators.add_property_validator('host', host_validator)
non_field_validators.add_property_validator('basePath', base_path_validator)
non_field_validators.add_property_validator('schemes', schemes_validator)
non_field_validators.add_property_validator('produces', mimetype_validator)
non_field_validators.add_property_validator('consumes', mimetype_validator)
non_field_validators.add_property_validator('paths', paths_validator)

swagger_schema_validator = generate_object_validator(
    schema=swagger_schema,
    non_field_validators=non_field_validators,
)



from flex.constants import (
    STRING,
)
from flex.validation.common import (
    generate_object_validator,
)


info_schema = {
    'required': [
        'title',
    ],
    'properties': {
        'title': {
            'type': STRING,
        },
        'description': {'type': STRING},
        'termsOfService': {'type': STRING},
        'contact': {'type': STRING},
        'license': {'type': STRING},
        'version': {'type': STRING},
    }
}

info_validator = generate_object_validator(
    schema=info_schema,
)



from flex.constants import (
    BOOLEAN,
)
from flex.validation.common import (
    generate_object_validator,
)
from flex.validation.schema import (
    construct_schema_validators,
)


read_only_schema = {
    'type': BOOLEAN,
}
read_only_validators = construct_schema_validators(read_only_schema, {})

read_only_validator = generate_object_validator(
    field_validators=read_only_validators,
)



from flex.error_messages import MESSAGES
from flex.exceptions import ValidationError
from flex.constants import (
    EMPTY,
    INTEGER,
    STRING,
)
from flex.utils import pluralize
from flex.validation.common import (
    generate_object_validator,
)
from flex.decorators import (
    pull_keys_from_obj,
    suffix_reserved_words,
    skip_if_any_kwargs_empty,
)


@pull_keys_from_obj('minLength', 'maxLength')
@skip_if_any_kwargs_empty('minLength', 'maxLength')
def validate_max_length_greater_than_or_equal_to_min_length(minLength, maxLength, **kwargs):
    if minLength is EMPTY or maxLength is EMPTY:
        return

    if not maxLength >= minLength:
        raise ValidationError(
            MESSAGES['max_length']['must_be_greater_than_min_length']
        )


@pull_keys_from_obj('type', 'maxLength')
@suffix_reserved_words
@skip_if_any_kwargs_empty('type_', 'maxLength')
def validate_type_for_max_length(type_, maxLength, **kwargs):
    types = pluralize(type_)

    if not set(types).intersection((STRING,)):
        raise ValidationError(
            MESSAGES['type']['invalid_type_for_max_length'],
        )


max_length_schema = {
    'type': INTEGER,
    'minimum': 1,
}
max_length_validator = generate_object_validator(
    schema=max_length_schema,
)



from flex.constants import (
    STRING,
)
from flex.validation.common import (
    generate_object_validator,
)


name_schema = {
    'type': STRING,
}


name_validator = generate_object_validator(
    schema=name_schema,
)



from flex.error_messages import MESSAGES
from flex.exceptions import ValidationError
from flex.constants import (
    ARRAY,
    INTEGER,
)
from flex.utils import pluralize
from flex.validation.common import (
    generate_object_validator,
)
from flex.validation.schema import (
    construct_schema_validators,
)
from flex.decorators import (
    pull_keys_from_obj,
    suffix_reserved_words,
    skip_if_any_kwargs_empty,
)


@pull_keys_from_obj('type', 'minItems')
@suffix_reserved_words
@skip_if_any_kwargs_empty('type_', 'minItems')
def validate_type_for_min_items(type_, minItems, **kwargs):
    types = pluralize(type_)

    if not set(types).intersection((ARRAY,)):
        raise ValidationError(
            MESSAGES['type']['invalid_type_for_min_items'],
        )


min_items_schema = {
    'type': INTEGER,
    'minimum': 0,
}
min_items_validators = construct_schema_validators(min_items_schema, {})

min_items_validator = generate_object_validator(
    field_validators=min_items_validators,
)



import itertools
import collections
import functools
import re

from flex.exceptions import MultiplePathsFound
from flex.error_messages import MESSAGES
from flex.constants import (
    PATH,
    REQUEST_METHODS,
)
from flex.parameters import (
    find_parameter,
    merge_parameter_lists,
    dereference_parameter_list,
)


REGEX_REPLACEMENTS = (
    ('\.', '\.'),
    ('\{', '\{'),
    ('\}', '\}'),
)


def escape_regex_special_chars(api_path):
    """
    Turns the non prametrized path components into strings subtable for using
    as a regex pattern.  This primarily involves escaping special characters so
    that the actual character is matched in the regex.
    """
    def substitute(string, replacements):
        pattern, repl = replacements
        return re.sub(pattern, repl, string)

    return functools.reduce(substitute, REGEX_REPLACEMENTS, api_path)


# matches the parametrized parts of a path.
# eg. /{id}/ matches the `{id}` part of it.
PARAMETER_REGEX = re.compile('(\{[^\}]+})')


def construct_parameter_pattern(parameter):
    """
    Given a parameter definition returns a regex pattern that will match that
    part of the path.
    """
    name = parameter['name']

    return "(?P<{name}>.+)".format(name=name)


def process_path_part(part, parameters):
    """
    Given a part of a path either:
        - If it is a parameter:
            parse it to a regex group
        - Otherwise:
            escape any special regex characters
    """
    if PARAMETER_REGEX.match(part):
        parameter_name = part.strip('{}')
        try:
            parameter = find_parameter(parameters, name=parameter_name, in_=PATH)
        except ValueError:
            pass
        else:
            return construct_parameter_pattern(parameter)
    return escape_regex_special_chars(part)


def get_parameter_names_from_path(api_path):
    return tuple(p.strip('{}') for p in PARAMETER_REGEX.findall(api_path))


def path_to_pattern(api_path, parameters):
    """
    Given an api path, possibly with parameter notation, return a pattern
    suitable for turing into a regular expression which will match request
    paths that conform to the parameter definitions and the api path.
    """
    parts = re.split(PARAMETER_REGEX, api_path)
    pattern = ''.join((process_path_part(part, parameters) for part in parts))

    if not pattern.startswith('^'):
        pattern = "^{0}".format(pattern)
    if not pattern.endswith('$'):
        pattern = "{0}$".format(pattern)

    return pattern


def path_to_regex(api_path, path_parameters, operation_parameters=None, global_parameters=None):
    if global_parameters is None:
        global_parameters = {}
    if operation_parameters is None:
        operation_parameters = []
    pattern = path_to_pattern(
        api_path=api_path,
        parameters=merge_parameter_lists(
            global_parameters.values(),
            dereference_parameter_list(path_parameters, global_parameters),
            dereference_parameter_list(operation_parameters, global_parameters),
        ),
    )
    return re.compile(pattern)


def extract_path_parameters(path_definition):
    if not path_definition:
        return []
    return path_definition.get('parameters', [])


def extract_operation_parameters(path_definition):
    return list(itertools.chain.from_iterable(
        (v or {}).get('parameters', [])
        for k, v in path_definition.items()
        if k in REQUEST_METHODS
    ))


def match_path_to_api_path(path_definitions, target_path, base_path='', global_parameters=None):
    """
    Match a request or response path to one of the api paths.

    Anything other than exactly one match is an error condition.
    """
    if global_parameters is None:
        global_parameters = {}
    assert isinstance(global_parameters, collections.Mapping)
    if target_path.startswith(base_path):
        target_path = target_path[len(base_path):]

    # Convert all of the api paths into Path instances for easier regex matching.
    paths = {
        p: path_to_regex(
            api_path=p,
            path_parameters=extract_path_parameters(v),
            operation_parameters=extract_operation_parameters(v),
            global_parameters=global_parameters,
        )
        for p, v in path_definitions.items()
    }

    matching_api_paths = [(p, r.match(target_path))
                          for p, r in paths.items() if r.match(target_path)]

    if not matching_api_paths:
        raise LookupError(MESSAGES['path']['no_matching_paths_found'].format(target_path))
    elif len(matching_api_paths) > 1:
        # TODO: This area needs improved logic.
        # We check to see if any of the matched paths is longers than
        # the others.  If so, we *assume* it is the correct match.  This is
        # going to be prone to false positives. in certain cases.
        matches_by_path_size = collections.defaultdict(list)
        for path, match in matching_api_paths:
            matches_by_path_size[len(path)].append(path)
        longest_match = max(matches_by_path_size.keys())
        if len(matches_by_path_size[longest_match]) == 1:
            return matches_by_path_size[longest_match][0]
        raise MultiplePathsFound(
            MESSAGES['path']['multiple_paths_found'].format(
                target_path, [v[0] for v in matching_api_paths],
            )
        )
    else:
        return matching_api_paths[0][0]



import collections

from flex.exceptions import (
    ValidationError,
    ErrorDict,
    ErrorList,
)
from flex.constants import (
    OBJECT,
)
from flex.utils import (
    is_non_string_iterable,
)
from flex.decorators import (
    skip_if_not_of_type,
    skip_if_empty,
)
from flex.functional import (
    apply_functions_to_key,
)


class ValidationList(list):
    def __init__(self, value=None):
        super(ValidationList, self).__init__()
        if value:
            self.add_validator(value)

    def add_validator(self, validator):
        if is_non_string_iterable(validator)\
           and not isinstance(validator, collections.Mapping):
            for value in validator:
                self.add_validator(value)
        else:
            self.append(validator)

    def validate_object(self, obj, **kwargs):
        with ErrorList() as errors:
            for validator in self:
                try:
                    validator(obj, **kwargs)
                except ValidationError as err:
                    errors.add_error(err.detail)

    def __call__(self, *args, **kwargs):
        return self.validate_object(*args, **kwargs)


class ValidationDict(collections.defaultdict):
    def __init__(self, validators=None):
        super(ValidationDict, self).__init__(ValidationList)
        if validators is not None:
            if not isinstance(validators, collections.Mapping):
                raise ValueError("ValidationDict may only be instantiated with a mapping")
            for key, validator in validators.items():
                self.add_validator(key, validator)

    def add_validator(self, key, validator):
        self[key].add_validator(validator)

    def add_property_validator(self, key, validator):
        self.add_validator(
            key,
            skip_if_empty(skip_if_not_of_type(OBJECT)(apply_functions_to_key(key, validator)))
        )

    def update(self, other):
        for key, value in other.items():
            self.add_validator(key, value)

    def validate_object(self, obj, **kwargs):
        with ErrorDict() as errors:
            for key, validator in self.items():
                try:
                    validator(obj, **kwargs)
                except ValidationError as err:
                    errors.add_error(key, err.detail)

    def __call__(self, *args, **kwargs):
        return self.validate_object(*args, **kwargs)



from flex.constants import (
    OBJECT,
    ARRAY,
)
from flex.datastructures import (
    ValidationDict,
)

from flex.loading.common.format import (
    format_validator,
)
from .title import (
    title_validator,
)
from flex.loading.common.default import (
    validate_default_is_of_one_of_declared_types,
)
from flex.loading.common.external_docs import (
    external_docs_validator,
)
from .min_properties import (
    min_properties_validator,
    validate_type_for_min_properties,
)
from .max_properties import (
    max_properties_validator,
    validate_max_properties_is_greater_than_or_equal_to_min_properties,
    validate_type_for_max_properties,
)
from .required import (
    required_validator,
)
from .type import (
    type_validator,
)
from .read_only import (
    read_only_validator,
)
from flex.loading.common import (
    field_validators as common_field_validators,
    non_field_validators as common_non_field_validators,
    type_validators as common_type_validators,
)

schema_schema = {
    'type': OBJECT,
}

schema_field_validators = ValidationDict()
schema_field_validators.update(common_field_validators)

schema_field_validators.add_property_validator('format', format_validator)
schema_field_validators.add_property_validator('title', title_validator)
schema_field_validators.add_property_validator('minProperties', min_properties_validator)
schema_field_validators.add_property_validator('maxProperties', max_properties_validator)
schema_field_validators.add_property_validator('required', required_validator)
schema_field_validators.add_property_validator('type', type_validator)
schema_field_validators.add_property_validator('readOnly', read_only_validator)
schema_field_validators.add_property_validator('externalDocs', external_docs_validator)

schema_non_field_validators = ValidationDict()
schema_non_field_validators.update(common_non_field_validators)
schema_non_field_validators.update(common_type_validators)


schema_non_field_validators.add_validator(
    'default', validate_default_is_of_one_of_declared_types,
)
schema_non_field_validators.add_validator(
    'maxProperties', validate_max_properties_is_greater_than_or_equal_to_min_properties,
)
schema_non_field_validators.add_validator(
    'type', validate_type_for_min_properties,
)
schema_non_field_validators.add_validator(
    'type', validate_type_for_max_properties,
)


#
# Properties.
#
properties_schema = {
    'type': OBJECT,
}


#
# Items
#
items_schema = {
    'type': [
        ARRAY,  # Array of schemas
        OBJECT,  # Single Schema
    ]
}



from flex.constants import (
    STRING,
)
from flex.validation.common import (
    generate_object_validator,
)


swagger_version_schema = {
    'enum': ['2.0'],
    'type': STRING,
}

swagger_version_validator = generate_object_validator(
    schema=swagger_version_schema,
)



from __future__ import unicode_literals


TYPE_MESSAGES = {
    'unknown': 'Unknown type: {0}',
    'invalid': "Got value `{0}` of type `{1}`.  Value must be of type(s): `{2}`",
    'invalid_header_type': (
        "Invalid type for header: `{0}`.  Must be one of 'string', 'number', "
        "'integer', 'boolean', or 'array'."
    ),
    'invalid_type_for_minimum': '`minimum` can only be used for json number types',
    'invalid_type_for_maximum': '`maximum` can only be used for json number types',
    'invalid_type_for_multiple_of': '`multipleOf` can only be used for json number types',
    'invalid_type_for_min_length': '`minLength` can only be used for string types',
    'invalid_type_for_max_length': '`maxLength` can only be used for string types',
    'invalid_type_for_min_items': '`minItems` can only be used for array types',
    'invalid_type_for_max_items': '`maxItems` can only be used for array types',
    'invalid_type_for_unique_items': '`uniqueItems` can only be used for array types',
    'invalid_type_for_min_properties': 'minProperties can only be used for `object` types',
    'invalid_type_for_max_properties': 'maxProperties can only be used for `object` types',
    'non_body_parameters_must_declare_a_type': (
        "A Parameter who's `in` value is not 'body' must declare a type."
    ),
}

FORMAT_MESSAGES = {
    'invalid': "Value {0} does not conform to the format {1}",
    'invalid_uuid': "{0} is not a valid uuid",
    'invalid_datetime': "{0} is not a valid iso8601 date-time",
    'too_many_bits': "Integer {0} has {1} bits.  Must be no more than {2} bits",
    'invalid_uri': "The value `{0}` is not valid according to RFC3987.",
    'invalid_email': "The email address `{0}` is invalid according to RFC5322.",
}

REQUIRED_MESSAGES = {
    'required': "This value is required",
    'path_parameters_must_be_required': (
        "A Parameter who's `in` value is 'path' must be declared as required."
    ),
}

MULTIPLE_OF_MESSAGES = {
    'invalid': "Value must be a multiple of {0}.  Got {1} which is not.",
}

MINIMUM_AND_MAXIMUM_MESSAGES = {
    'invalid': "{0} must be {1} than {2}",
    'must_be_greater_than_minimum': (
        "The value of `maximum` must be greater than or equal to the value of `minimum`"
    ),
    'exclusive_minimum_required_minimum': (
        "When `exclusiveMinimum` is set, `minimum` is required"
    ),
    'exclusive_maximum_required_maximum': (
        "When `exclusiveMaximum` is set, `maximum` is required"
    ),
}


MIN_LENGTH_MESSAGES = {
    "invalid": "value must be no less than {0} characters in length.",
}


MAX_LENGTH_MESSAGES = {
    "invalid": "value must be no greater than {0} characters in length.",
    'must_be_greater_than_min_length': (
        'The value of `maxLength` must be greater than or equal to the `minLength` value'
    ),
}


MIN_ITEMS_MESSAGES = {
    'invalid': "Array must have at least {0} items.  It had only had {1} items.",
}


MAX_ITEMS_MESSAGES = {
    'must_be_greater_than_min_items': (
        "The value of `maxItems` must be greater than or equal to the value of `minItems`"
    ),
    'invalid': "Array must have no more than {0} items.  It had {1} items.",
}


UNIQUE_ITEMS_MESSAGES = {
    'invalid': "Array items must be unique.  The following items appeard more than once: {0}",
}


ENUM_MESSAGES = {
    'invalid': "Invalid value.  {0} is not one of the available options ({1})",
}


PATTERN_MESSAGES = {
    'invalid_regex': "{0} is not a valid regular expression",
    'invalid': "{0} did not match the pattern `{1}`.",
}


MIN_PROPERTIES_MESSAGES = {
    'invalid': "Object must have more than {0} properties.  It had {1}",
}


MAX_PROPERTIES_MESSAGES = {
    'invalid': "Object must have less than {0} properties.  It had {1}",
    'must_be_greater_than_min_properties': (
        "The value of `maxProperties` must be greater than or equal to `minProperties`."
    ),
}


ITEMS_MESSAGES = {
    'invalid_type': '`items` must be a reference, a schema, or an array of schemas.',
    'items_required_for_type_array': (
        "For type \"array\", the items is required."
    )
}


DEFAULT_MESSAGES = {
    'invalid_type': (
        "The value of `default` must be of one of the declared types for the "
        "schema.  `{0}` is not one of `{1}`"
    ),
}


REQUEST_MESSAGES = {
    'invalid_method': (
        'Request was not one of the allowed request methods.  Got '
        '`{0}`: Expected one of: `{1}`'
    ),
}


RESPONSE_MESSAGES = {
    'invalid_status_code': (
        "Request status code was not found in the known response codes.  Got "
        "`{0}`: Expected one of: `{1}`"
    )
}


PATH_MESSAGES = {
    'no_matching_paths_found': 'No paths found for {0}',
    'multiple_paths_found': (
        'Unable to determine path for {0}.  Found multiple matches: `{1}`'
    ),
    'unknown_path': 'Request path did not match any of the known api paths.',
    'missing_parameter': (
        "The parameter named `{0}` is declared to be a PATH parameter but does "
        "not appear in the api path `{1}`.  All path parameters must exist as a "
        "parameter in the api path"
    ),
    'must_start_with_slash': "Path must start with a '/'",
    'invalid': "Invalid Path: {0}",
}


UNKNOWN_REFERENCE_MESSAGES = {
    'security': "Unknown SecurityScheme reference `{0}`",
    'parameter': "Unknown Parameter reference `{0}`",
    'definition': 'Unknown definition reference `{0}`',
    'no_definitions': "No definitions found in context",
}


CONTENT_TYPE_MESSAGES = {
    'invalid': 'Invalid content type `{0}`.  Must be one of `{1}`.',
}


HOST_MESSAGES = {
    'invalid': (
        "Invalid host: {0}. This MUST be the host only and does not include the "
        "scheme nor sub-paths. It MAY include a port. If the host is not "
        "included, the host serving the documentation is to be used (including "
        "the port)"
    ),
    'may_not_include_path': (
        "Invalid host: {0}.  Includes the path component.  The host value "
        "should be the host only, without path or scheme."
    ),
    'may_not_include_scheme': (
        "Invalid host: {0}.  Includes the scheme component.  The host value "
        "should be the host only, without path or scheme."
    ),
}

SCHEMES_MESSAGES = {
    'invalid': "Invalid scheme: {0}. Must be one of (http, https, ws, wss).",
}


MIMETYPE_MESSAGES = {
    'invalid': "Invalid mimetype: {0}.",
}


REFERENCE_MESSAGES = {
    'undefined': "The $ref `{0}` is referenced but never defined",
}


SCHEMA_MESSAGES = {
    'body_parameters_must_include_a_schema': (
        "A Parameter who's `in` value is 'body' must declare a schema."
    ),
}

COLLECTION_FORMAT_MESSAGES = {
    'invalid_based_on_in_value': (
        "The collectionFormat 'multi' is only valid for `in` values of "
        "\"query\" or \"formData\"."
    ),
}


MESSAGES = {
    'type': TYPE_MESSAGES,
    'format': FORMAT_MESSAGES,
    'required': REQUIRED_MESSAGES,
    'multiple_of': MULTIPLE_OF_MESSAGES,
    'minimum': MINIMUM_AND_MAXIMUM_MESSAGES,
    'maximum': MINIMUM_AND_MAXIMUM_MESSAGES,
    'max_length': MAX_LENGTH_MESSAGES,
    'min_length': MIN_LENGTH_MESSAGES,
    'min_items': MIN_ITEMS_MESSAGES,
    'max_items': MAX_ITEMS_MESSAGES,
    'min_properties': MIN_PROPERTIES_MESSAGES,
    'max_properties': MAX_PROPERTIES_MESSAGES,
    'unique_items': UNIQUE_ITEMS_MESSAGES,
    'enum': ENUM_MESSAGES,
    'pattern': PATTERN_MESSAGES,
    'items': ITEMS_MESSAGES,
    'request': REQUEST_MESSAGES,
    'response': RESPONSE_MESSAGES,
    'path': PATH_MESSAGES,
    'unknown_reference': UNKNOWN_REFERENCE_MESSAGES,
    'content_type': CONTENT_TYPE_MESSAGES,
    'host': HOST_MESSAGES,
    'schemes': SCHEMES_MESSAGES,
    'mimetype': MIMETYPE_MESSAGES,
    'default': DEFAULT_MESSAGES,
    'reference': REFERENCE_MESSAGES,
    'schema': SCHEMA_MESSAGES,
    'collection_format': COLLECTION_FORMAT_MESSAGES,
}



from flex.constants import (
    STRING,
)
from flex.validation.common import (
    generate_object_validator,
)
from flex.validation.schema import (
    construct_schema_validators,
)


format_schema = {
    'type': STRING,
}
format_validators = construct_schema_validators(format_schema, {})

format_validator = generate_object_validator(
    field_validators=format_validators,
)



from flex.constants import (
    STRING,
)
from flex.validation.common import (
    generate_object_validator,
)

description_schema = {
    'type': STRING,
}


description_validator = generate_object_validator(
    schema=description_schema,
)



import re

from flex.exceptions import ValidationError
from flex.constants import (
    STRING,
)
from flex.error_messages import MESSAGES
from flex.validation.common import (
    generate_object_validator,
)
from flex.validation.schema import (
    construct_schema_validators,
)
from flex.decorators import (
    skip_if_empty,
    skip_if_not_of_type,
)


@skip_if_empty
@skip_if_not_of_type(STRING)
def regex_validator(value, **kwargs):
    try:
        re.compile(value)
    except re.error:
        raise ValidationError(
            MESSAGES['pattern']['invalid_regex'].format(value)
        )

pattern_schema = {
    'type': STRING,
}
pattern_validators = construct_schema_validators(pattern_schema, {})
pattern_validators.add_validator('regex', regex_validator)

pattern_validator = generate_object_validator(
    field_validators=pattern_validators,
)



from flex.constants import (
    STRING,
)
from flex.validation.common import (
    generate_object_validator,
)
from flex.validation.schema import (
    construct_schema_validators,
)


title_schema = {
    'type': STRING,
}
title_validators = construct_schema_validators(title_schema, {})

title_validator = generate_object_validator(
    field_validators=title_validators,
)



import re

from flex.datastructures import (
    ValidationList,
)
from flex.constants import (
    ARRAY,
)
from flex.exceptions import ValidationError
from flex.error_messages import MESSAGES
from flex.validation.common import (
    generate_object_validator,
)
from flex.decorators import (
    skip_if_empty,
    skip_if_not_of_type,
)


# top-level type name / [ tree. ] subtype name [ +suffix ] [ ; parameters ]
MIMETYPE_PATTERN = (
    '^'
    '(application|audio|example|image|message|model|multipart|text|video)'  # top-level type name
    '/'
    '(vnd(\.[-a-zA-Z0-9]+)*\.)?'  # vendor tree
    '([-a-zA-Z0-9]+)'  # media type
    '(\+(xml|json|ber|der|fastinfoset|wbxml|zip))?'
    '((; [-a-zA-Z0-9]+=(([-\.a-zA-Z0-9]+)|(("|\')[-\.a-zA-Z0-9]+("|\'))))+)?'  # parameters
    '$'
)


@skip_if_empty
@skip_if_not_of_type(ARRAY)
def validate_mimetype(values, **kwargs):
    for value in values:
        if not re.match(MIMETYPE_PATTERN, value):
            raise ValidationError(
                MESSAGES['mimetype']['invalid'].format(value),
            )


mimetype_schema = {
    'type': ARRAY,
}

non_field_validators = ValidationList()
non_field_validators.add_validator(validate_mimetype)

mimetype_validator = generate_object_validator(
    schema=mimetype_schema,
    non_field_validators=non_field_validators,
)
